import BookTrackerPlugin from "@src/main";
import { App } from "obsidian";

export class Storage {
	public constructor(
		private readonly app: App,
		private readonly plugin: BookTrackerPlugin
	) {}

	private getFilePath(filename: string): string {
		return `${this.plugin.manifest.dir!!}/${filename}`;
	}

	public async readJSON<T>(filename: string): Promise<T | null> {
		const filePath = this.getFilePath(filename);
		const content = await this.app.vault.adapter.read(filePath);
		if (!content) {
			return null;
		}

		try {
			return JSON.parse(content) as T;
		} catch (error) {
			console.error(`Error parsing JSON from ${filePath}:`, error);
			return null;
		}
	}

	public async writeJSON<T>(filename: string, data: T): Promise<void> {
		const filePath = this.getFilePath(filename);
		const content = JSON.stringify(data, null, 2);
		await this.app.vault.adapter.write(filePath, content);
	}
}

interface ReadingLogEntry {
	readonly book: string;
	readonly pagesRead: number;
	readonly pagesReadTotal: number;
	readonly createdAt: Date;
}

export class ReadingLog {
	private entries: ReadingLogEntry[] = [];

	public constructor(
		private readonly app: App,
		private readonly plugin: BookTrackerPlugin
	) {
		this.loadEntries().catch((error) => {
			console.error("Failed to load reading log entries:", error);
		});
	}

	private async loadEntries() {
		const entries = await this.plugin.storage.readJSON<ReadingLogEntry[]>(
			"reading-log.json"
		);
		if (entries) {
			this.entries = entries;
		}
	}

	private async storeEntries() {
		await this.plugin.storage.writeJSON("reading-log.json", this.entries);
	}

	public getLatestEntry(book: string): ReadingLogEntry | null {
		const entriesForBook = this.entries.filter(
			(entry) => entry.book === book
		);

		return entriesForBook.length > 0
			? entriesForBook[entriesForBook.length - 1]
			: null;
	}

	public async addEntry(book: string, pageEnded: number): Promise<void> {
		const latestEntry = this.getLatestEntry(book);

		const newEntry: ReadingLogEntry = {
			book,
			pagesRead: latestEntry
				? pageEnded - latestEntry.pagesReadTotal
				: pageEnded,
			pagesReadTotal: pageEnded,
			createdAt: new Date(),
		};

		this.entries.push(newEntry);
		await this.storeEntries();
	}

	public async removeEntries(book: string): Promise<void> {
		this.entries = this.entries.filter((entry) => entry.book !== book);
		await this.storeEntries();
	}

	public async removeLastEntry(book: string): Promise<void> {
		const latestEntryIndex = this.entries.findLastIndex(
			(entry) => entry.book === book
		);

		if (latestEntryIndex !== -1) {
			this.entries.splice(latestEntryIndex, 1);
			await this.storeEntries();
		}
	}

	public async clearEntries(): Promise<void> {
		this.entries = [];
		await this.storeEntries();
	}
}
